/* See {sample_conv.h}. */
/* Last edited on 2007-10-10 13:26:30 by stolfi */

#include <sample_conv.h>

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <values.h>
#include <math.h>

#include <affirm.h>
#include <bool.h>

float sample_conv_floatize
  ( sample_uint_t iv, 
    sample_uint_t maxval, 
    double lo,
    double hi, 
    sample_uint_t *imin, 
    sample_uint_t *imax, 
    float *vmin,
    float *vmax
  )
  /* Convert integer intensity to float, kep stats: */
  { if (iv < (*imin)) { (*imin) = iv; }
    if (iv > (*imax)) { (*imax) = iv; }
    double rv = ((double)iv + 0.5)/((double)maxval + 1.0);
    float fv = lo + rv*(hi - lo);
    if (fv < (*vmin)) { (*vmin) = fv; }
    if (fv > (*vmax)) { (*vmax) = fv; }
    return fv;
  }

void sample_conv_print_floatize_stats
  ( int iChan,           /* Channel index in input image. */
    int oChan,           /* Channel index in output image. */
    sample_uint_t imin,   /* Minimum integer sample seen. */
    sample_uint_t imax,   /* Maximum integer sample seen. */
    sample_uint_t maxval, /* Maximum possible integer sample. */
    double lo,           /* Low end of float scaling range. */
    double hi,           /* High end of float scaling range. */
    float vmin,          /* Minimum float sample seen. */
    float vmax           /* Maximum float sample seen. */
  )
  { fprintf(stderr, "  converted int channel %d to float channel %d:\n", iChan, oChan);
    fprintf(stderr, "    mapped [ 0 .. %u ] to [ %12.5e _ %12.5e ]\n", maxval, lo, hi);
    fprintf(stderr, "    actual input range  = [ %5u .. %5u ]\n", imin, imax);
    fprintf(stderr, "    actual output range = [ %12.5e _ %12.5e]\n", vmin, vmax);
  }

sample_uint_t sample_conv_quantize
  ( float fv, 
    sample_uint_t maxval, 
    double lo,
    double hi, 
    float *vmin,
    float *vmax, 
    int *clo,
    int *chi,
    sample_uint_t *imin, 
    sample_uint_t *imax
  )
  { if (fv < (*vmin)) { (*vmin) = fv; }
    if (fv > (*vmax)) { (*vmax) = fv; }
    double fdelta = hi - lo;
    double rv = (fdelta == 0 ? 0.5 : (fv - lo)/fdelta);
    if (rv < 0.0) { rv = 0.0; (*clo)++; }
    if (rv > 1.0) { rv = 1.0; (*chi)++; }
    sample_uint_t zv = (rv == 1.0 ? maxval : (int)floor(rv*(maxval + 1)));
    demand((zv >= 0) && (zv <= (int)maxval), "bad zv");
    sample_uint_t iv = zv;
    if (iv < (*imin)) { (*imin) = iv; }
    if (iv > (*imax)) { (*imax) = iv; }
    return iv;
  }
  
void sample_conv_print_quantize_stats
  ( int iChan,           /* Channel index in input image. */
    int oChan,           /* Channel index in output image. */
    float vmin,          /* Minimum float sample seen. */
    float vmax,          /* Maximum float sample seen. */
    double lo,           /* Low end of float scaling range. */
    double hi,           /* High end of float scaling range. */
    int clo,             /* Number of samples seen below {lo}. */
    int chi,             /* Number of samples seen above {hi}. */
    sample_uint_t maxval, /* Maximum possible integer sample. */
    sample_uint_t imin,   /* Minimum integer sample seen. */
    sample_uint_t imax    /* Maximum integer sample seen. */
  )
  { fprintf(stderr, "  converted float channel %d to int channel %d:\n", iChan, oChan);
    fprintf(stderr, "    mapped [ %12.5e _ %12.5e ] to [ 0 .. %u ]\n", lo, hi, maxval);
    fprintf(stderr, "    actual input range  = [ %12.5e _ %12.5e ]\n", vmin, vmax);
    fprintf(stderr, "    actual output range = [ %5u .. %5u ]\n", imin, imax);
    if ((clo > 0) || (chi > 0)) 
      { fprintf(stderr, "    clipped pixels: too low = %d  too high = %d\n", clo, chi); }
  }

