/* See {dg_plot_slow.h}. */
/* Last edited on 2007-10-15 00:21:07 by stolfi */

#define dg_plot_slow_C_COPYRIGHT "Copyright  2007 by the State University of Campinas (UNICAMP)."

#define _GNU_SOURCE
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <math.h>

#include <dg_plot_slow.h>

#include <bool.h>
#include <jsmath.h>
#include <affirm.h>
#include <rn.h>
#include <pswr.h>
#include <pswr_color.h>
#include <pswr_iso.h>

#include <dg_grid.h>
#include <dg_spline.h>
#include <dg_plot.h>
#include <dg_makegrid.h>

#include <bz_basic.h>

#define DDIM (2)
  /* Domain dimension. */

#define MAX_VALUES (3)
  /* Max number of function values (excluding shape). */

#define MAX_RDIM (DDIM + MAX_VALUES)
  /* Max number of function values (including shape). */

#define NFACES (9)
  /* Number of faces in a cell of dimension {DDIM}, namely {3^DDIM}. */

/* INTERNAL PROTOTYPES */

void dg_plot_slow_vertex(PSStream *ps, bz_patch_t *b, int steps);
void dg_plot_slow_edge(PSStream *ps, bz_patch_t *b, int steps);
void dg_plot_slow_face(PSStream *ps, bz_patch_t *b, int steps, double Rf, double Gf, double Bf);
     
bz_patch_t *dg_plot_slow_extract_face(bz_patch_t *b, box_face_index_t f);
  /* Extracts face {f} of the Bezier patch {b} and stores it into a newly
    allocated record. */

/* IMPLEMENTATIONS */

void dg_plot_slow(PSStream *ps, bz_patch_t *bz)
  { 
    demand(bz->m == 2, "wrong domain dimension");
    demand(bz->n >= 2, "bad range dimension");
    demand(bz->n <= MAX_RDIM, "bad range dimension");
    int BSTEPS = 10;
    double Rf = 1.0, Gf = 0.9, Bf = 0.6;
    int fix;
    fprintf(stderr, "+dg_plot_slow dim = %d\n", bz->m);
    /* Plot faces */
    for (fix = 0; fix < NFACES; fix++)
      { if (box_face_dimension(fix, bz->m) == 2)
        { bz_patch_t *t = dg_plot_slow_extract_face(bz, fix);
          dg_plot_slow_face(ps, t, BSTEPS, Rf, Gf, Bf);
          free(t->c); free(t);
        }
      }
    /* Plot edges */
    pswr_set_pen(ps, 0.0, 0.0, 0.0,   0.15, 0.0, 0.0);
    for (fix = 0; fix < NFACES; fix++)
      { if (box_face_dimension(fix, bz->m) == 1)
        { bz_patch_t *t = dg_plot_slow_extract_face(bz, fix);
          dg_plot_slow_edge(ps, t, BSTEPS);
          free(t->c); free(t);
        }
      }
    /* Plot vertices */
    pswr_set_pen(ps, 0.0, 0.0, 0.0,   0.15, 0.0, 0.0);
    for (fix = 0; fix < NFACES; fix++)
      if (box_face_dimension(fix, bz->m) == 0) 
        { bz_patch_t *t = dg_plot_slow_extract_face(bz, fix);
          dg_plot_slow_vertex(ps, t, BSTEPS);
          free(t->c); free(t);
        }
    fprintf(stderr, "-dg_plot_slow\n");
  }

void dg_plot_slow_vertex(PSStream *ps, bz_patch_t *b, int steps)
  {
    fprintf(stderr, "+dg_plot_slow_vertex dim = %d\n", b->m);
    affirm(b->m == 0, "invalid vertex patch");
    pswr_set_fill_color(ps, 0.0,0.0,0.0);
    pswr_dot(ps, b->c[0], b->c[1], 0.25, TRUE, FALSE);
    fprintf(stderr, "-dg_plot_slow_vertex\n");
  }

void dg_plot_slow_edge(PSStream *ps, bz_patch_t *b, int steps)
  {
    fprintf(stderr, "+dg_plot_slow_edge dim = %d\n", b->m);
    affirm(b->m == 1, "invalid edge patch");
    double p[2], q[2];
    int i;
    fprintf(stderr, "shape =\n");
    bz_patch_print(stderr, b, "%6.2f");
    for (i = 0; i <= steps; i++)
      { double ui = ((double)i)/((double)steps);
        bz_patch_eval(b, &ui, p);
        if (i > 0) { pswr_segment(ps, q[0], q[1], p[0], p[1]); }
        q[0] = p[0]; q[1] = p[1];
      }
    fprintf(stderr, "-dg_plot_slow_edge\n");
  }

void dg_plot_slow_face(PSStream *ps, bz_patch_t *b, int steps, double Rf, double Gf, double Bf)
  {
    int i, j;
    fprintf(stderr, "+dg_plot_slow_face dim = %d\n", b->m);
    affirm(b->m == 2, "invalid face patch");
    fprintf(stderr, "shape =\n");
    bz_patch_print(stderr, b, "%6.2f");
    double u[2], v[2], p[MAX_RDIM], q[MAX_RDIM], r[MAX_RDIM], s[MAX_RDIM];
    q[0] =  q[1] = s[1] = s[0] = 0; /* To pacify the compiler. */
    for (i = 1; i <= steps; i++)
      { for (j = 0; j <= steps; j++)
          { u[0] = ((double)i)/((double)steps);
            u[1] = ((double)j)/((double)steps);
            bz_patch_eval(b, u, p);
            v[0] = ((double)i-1)/((double)steps);
            v[1] = ((double)j)/((double)steps);
            bz_patch_eval(b, v, r);
            if (j > 0) 
              { double xm = (p[0]+q[0]+r[0]+s[0])/4.0;
                double ym = (p[1]+q[1]+r[1]+s[1])/4.0;
                pswr_set_fill_color(ps, Rf,Gf,Bf);
                pswr_triangle(ps, p[0], p[1], q[0], q[1], xm, ym, TRUE, FALSE);
                pswr_triangle(ps, p[0], p[1], r[0], r[1], xm, ym, TRUE, FALSE);
                pswr_triangle(ps, s[0], s[1], q[0], q[1], xm, ym, TRUE, FALSE);
                pswr_triangle(ps, s[0], s[1], r[0], r[1], xm, ym, TRUE, FALSE);
              }
            q[0] = p[0]; q[1] = p[1];
            s[0] = r[0]; s[1] = r[1];
          }
      }
    fprintf(stderr, "-dg_plot_slow_face\n");
  }

bz_patch_t *dg_plot_slow_extract_face(bz_patch_t *b, box_face_index_t fix)
  {
    dg_dim_t bm = b->m, tm = 0;
    box_signed_dir_t loc[DDIM];
    box_face_signature(fix, bm, loc);
    bz_patch_t *t = (bz_patch_t *)notnull(malloc(sizeof(bz_patch_t)), "no mem");
    bz_degree_t tg[DDIM];
    int i;
    /* Extract degree sequence: */
    for(i = 0; i < bm; i++)
      { if (loc[i] == box_ZER) { tg[tm] = b->g[i]; tm++; } }
    (*t) = bz_patch_new(tm, b->n, tg);
    bz_patch_get_face(b, loc, t);
    return t;
  }

